from typing import Literal

# https://developer.mozilla.org/en-US/docs/Glossary/Void_element
VOID_ELEMENTS = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
]

ELEMENT_SIZE = Literal["sm", "md", "lg", "xl", "2xl"]

ELEMENT_TYPES = Literal[
    "a",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "bdi",
    "bdo",
    "blockquote",
    "body",
    "br",
    "body",
    "button",
    "canvas",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "data",
    "datalist",
    "dd",
    "details",
    "dfn",
    "dialog",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hgroup",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "input",
    "ins",
    "kbd",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "map",
    "mark",
    "menu",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "param",
    "picture",
    "pre",
    "progress",
    "q",
    "rq",
    "rt",
    "ruby",
    "s",
    "samp",
    "script",
    "search",
    "section",
    "select",
    "small",
    "source",
    "span",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "svg",
    "table",
    "tbody",
    "td",
    "template",
    "textarea",
    "tfoot",
    "th",
    "thread",
    "time",
    "title",
    "tr",
    "track",
    "u",
    "ul",
    "var",
    "video",
    "wbr",
]



from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.extensions.on_event import OnEvent


class Dropdown(OnEvent):
    root_class: str = "select "
    root_size: str = "select-{size}"
    _classes: str = "select-bordered w-full max-w-xs"

    def __init__(self, name: str, items: list[str], placeholder: Optional[str] = None) -> None:
        """Dropdown

        A dropdown is a list in which the selected item is always visible, and the others are visible on demand.

        .. code-block:: python
            from uiwiz import ui

            ui.dropdown("dropdown", ["Option 1", "Option 2", "Option 3"], "Select an option")
            ui.dropdown("dropdown", ["Option 1", "Option 2", "Option 3"], "Option 1") # Option 1 is selected by default
        
        :param name: The name of the dropdown
        :param items: List of items in the dropdown
        :param placeholder: Placeholder text
        """
        super().__init__("select")
        self.attributes["name"] = name
        self.classes()
        self.size(self._size)
        with self:
            if placeholder and placeholder not in items:
                Element("option disabled selected", content=placeholder)
            for v in items:
                e = Element("option", content=v)
                e.value = v
                if placeholder == v:
                    e.attributes["selected"] = "selected"



from uiwiz.elements.extensions.on_event import OnEvent


class Radio(OnEvent):
    root_class: str = "radio"
    root_size: str = "radio-{size}"

    def __init__(self, name: str, checked: bool = False) -> None:
        """Radio

        This element is used for radio buttons

        :param name: name of the radio group
        :param checked: if the radio is checked
        """
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "radio"
        if checked:
            self.attributes["checked"] = "checked"



from typing import Optional

from uiwiz.elements.extensions.on_event import OnEvent


class Range(OnEvent):
    root_class: str = "range "
    root_size: str = "range-{size}"

    def __init__(self, name: str, value: int, min: int, max: int, step: Optional[int] = None) -> None:
        """Range
        
        This element is used for range inputs

        .. code-block:: python
            from uiwiz import ui

            ui.range("range", 50, 0, 100, 1):

        :param name: name of the range input
        :param value: default value of the range input
        :param min: minimum value of the range input
        :param max: maximum value of the range input
        :param step: step value of the range input
        """
        super().__init__("input")
        self.attributes["type"] = "range"
        self.attributes["name"] = name
        self.attributes["min"] = min
        self.attributes["max"] = max
        self.attributes["value"] = value
        if step:
            self.attributes["step"] = step

    @property
    def min(self) -> int:
        return self.attributes["min"]

    @min.setter
    def min(self, value: int):
        self.attributes["min"] = value
        return self

    @property
    def max(self) -> int:
        return self.attributes["max"]

    @max.setter
    def max(self, value: int):
        self.attributes["max"] = value
        return self



import json
import logging
from html import escape
from mimetypes import guess_type
from pathlib import Path
from typing import Optional, Union

from fastapi import FastAPI, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import HTMLResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from jinja2 import Template
from starlette.requests import Request

from uiwiz.asgi_request_middleware import AsgiRequestMiddleware
from uiwiz.element import Element
from uiwiz.elements.button import Button
from uiwiz.elements.col import Col
from uiwiz.elements.html import Html
from uiwiz.frame import Frame
from uiwiz.page_route import PageRouter
from uiwiz.shared import register_path, resources
from uiwiz.static_middleware import AsgiTtlMiddleware
from uiwiz.version import __version__

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class CustomList(list):
    def append(self, route: APIRoute):
        if getattr(route, "endpoint", None):
            register_path(route.path, route.endpoint)
        return super().append(route)


class UiwizApp(FastAPI):
    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert alert-error",
        cache_age: int = 14400,
        theme: Optional[str] = None,
        title: Optional[str] = "UiWiz",
        auto_close_toast_error: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """App used for asgi applications

        See fastapi documentation for more *args and **kwargs
        
        :param toast_delay: The time in milliseconds before the toast is removed
        :param error_classes: The classes to apply to the toast error for default validation errors
        :param cache_age: The time in seconds to cache the static files
        :param theme: The default theme to use
        :param title: The default title for the app
        :param auto_close_toast_error: If the toast error should auto close
        :param args: FastAPI args
        :param kwargs: FastAPI kwargs
        """
        super().__init__(*args, **kwargs)
        self.router.routes = CustomList()
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        self.auto_close_toast_error = auto_close_toast_error
        if theme:
            self.theme = f"data-theme={theme}"
        else:
            self.theme = theme
        self.title = title
        self.templates = Jinja2Templates(Path(__file__).parent / "templates")
        self.add_static_files(f"/_static/{__version__}/", Path(__file__).parent / "static")

        self.add_middleware(AsgiRequestMiddleware)
        self.add_middleware(GZipMiddleware)
        self.add_middleware(AsgiTtlMiddleware, cache_age=cache_age)
        self.extensions: dict[str, Path] = {}
        self.app_paths: dict[str, Path] = {}

        @self.get(f"/_uiwiz/{__version__}/default.js", include_in_schema=False)
        def return_default_js(request: Request, response: Response):
            response.headers["cache-control"] = f"max-age={cache_age}"
            response.headers["x-uiwiz-content"] = "assets"
            return self.render(request, response, template_name="default.js", media_type="application/javascript")

        self.exception_handler(RequestValidationError)(self.handle_validation_error)
        

        @self.get("/_static/extension/{__version__}/{extension}/{filename}", include_in_schema=False)
        def get_extension(extension: str, filename: str):
            resource_key = f"{extension}/{filename}"
            if resource_key not in resources:
                return Response(status_code=404)

            with open(resources[resource_key], encoding="utf-8") as f:
                content = f.read()

            content_type, _ = guess_type(resource_key)
            return Response(content, media_type=content_type)

    def __get_title__(self, frame: Frame, route_title: Optional[str] = None) -> str:
        title = self.title

        if route_title:
            title = route_title

        if dynamic_title := frame.title:
            title = dynamic_title

        return title

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def page(
        self, path: str, *args, title: Optional[str] = None, favicon: Optional[str] = None, **kwargs
    ) -> PageRouter:
        return PageRouter().page(path, *args, title=title, favicon=favicon, router=self.router, **kwargs)

    def ui(self, path: str, *args, include_js: bool = True, include_css: bool = True, **kwargs) -> PageRouter:
        return PageRouter().ui(path=path, include_js=include_js, include_css=include_css, router=self.router, **kwargs)

    def render(
        self,
        request: Request,
        response: Response,
        title: Optional[str] = None,
        status_code: int = 200,
        template_name: str = "default.html",
        media_type: str = "text/html",
        root_overflow: str = "overflow-y: scroll",
    ):
        frame = Frame.get_stack()
        ext_js, ext_css = frame.render_ext()

        theme = self.theme
        if cookie_theme := request.cookies.get("data-theme"):
            theme = f"data-theme={escape(cookie_theme)}"

        root_overflow = f'style="{root_overflow}"'
        standard_headers = {"cache-control": "no-store", "x-uiwiz-content": "page"}
        default_page: Template = self.templates.get_template(template_name)

        if response:
            for key, value in response.headers.items():
                standard_headers[key] = value

        return self.return_funtion_response(
            HTMLResponse(
                content=default_page.render(
                    request=request,
                    root_element=[frame.render()],
                    title=self.__get_title__(frame, title),
                    theme=theme,
                    ext_js=ext_js,
                    ext_css=ext_css,
                    toast_delay=json.dumps({"delay": self.toast_delay}),
                    description_content=frame.meta_description_content,
                    overflow=root_overflow,
                    head=frame.head_ext,
                    version=__version__,
                ),
                status_code=status_code,
                headers=standard_headers,
                media_type=media_type,
            )
        )

    def return_funtion_response(self, response: Union[str, Response]) -> Union[str, Response]:
        Frame.get_stack().del_stack()
        return response

    async def handle_validation_error(self, request: Request, exc: RequestValidationError):
        fields_with_errors = [item.get("loc")[1] for item in exc.errors() if item.get("loc")[1] in exc.body]
        ok_fields = [item for item in exc.body.keys() if item not in fields_with_errors]

        Frame.get_stack().del_stack()
        Frame.get_stack()
        
        with Element().classes(self.error_classes) as toast:
            toast.attributes["id"] = "toast"
            toast.attributes["hx-swap-oob"] = "afterbegin"
            toast.attributes["hx-toast-data"] = json.dumps(jsonable_encoder(
                {
                    "detail": exc.errors(),
                    "fieldErrors": fields_with_errors,
                    "fieldOk": ok_fields,
                }
            ))
            html = Html("").classes("alert alert-error relative")
            html.tag = "span"
            html.attributes["hx-toast-data"] = json.dumps({"autoClose": self.auto_close_toast_error})
            html.attributes["hx-toast-delete-button"] = lambda: btn.id
            with html:
                with Col(gap="").classes("relative"):
                    for item in exc.errors():
                        Element(content=f"{item.get('loc')[1]}: {item.get('msg')}")
                if not self.auto_close_toast_error:
                    btn = Button("✕").classes("btn btn-sm btn-circle btn-ghost absolute right-2 top-2")
                    
        html_content = Frame.get_stack().render()


        return HTMLResponse(
            content=html_content,
            status_code=200,
            headers={"cache-control": "no-store", "x-uiwiz-content": "page", "x-uiwiz-validation-error": "true"},
        )


from uiwiz.elements.extensions.on_event import OnEvent


class Toggle(OnEvent):
    root_class: str = "toggle"
    root_size: str = "toggle-{size}"

    def __init__(self, name: str, checked: bool = False) -> None:
        """Toggle
        
        This element is used for toggle inputs

        .. code-block:: python
            from uiwiz import ui

            ui.toggle("toggle", True):

        :param name: name of the toggle input
        :param checked: default checked state of the toggle input
        """
        super().__init__("input")
        if checked:
            self.attributes["checked"] = "checked"
        self.attributes["type"] = "checkbox"
        self.attributes["name"] = name



from uiwiz.element import Element


class Avatar(Element):
    root_class: str = "avatar"
    _classes_inner: str = "w-{size} rounded-full"

    def __init__(self, path: str, size: int = 12) -> None:
        """
        Display an avatar image.

        :param path: The path to the image to display.
        :param size: The size of the avatar.
        """
        super().__init__()
        with self:
            with Element("div").classes(Avatar._classes_inner.format(size=size)) as container:
                self.container = container
                img = Element("img")
                img.attributes["src"] = path

    def classes(self, input: str = ""):
        if hasattr(self, "container"):
            self.container.classes(input)
        return self



from typing import Optional

from uiwiz.elements.extensions.on_event import OnEvent
from uiwiz.event import FUNC_TYPE, SWAP_EVENTS, TARGET_TYPE


class Button(OnEvent):
    root_class: str = "btn"
    root_size: str = "btn-{size}"

    def __init__(self, title: str) -> None:
        """
        Create a button element, with the given title.
        Can be used to trigger events.
        
        :param title: The title of the button.
        """
        super().__init__(tag="button")
        self.content = title

    def on_click(
        self,
        func: FUNC_TYPE = None,
        target: TARGET_TYPE = None,
        swap: SWAP_EVENTS = None,
        params: Optional[dict[str, str]] = None,
    ) -> "Button":
        self.event = {"func": func, "trigger": "click", "target": target, "swap": swap, "params": params}
        return self



import html
import json
from enum import Enum
from pathlib import Path
from typing import Any, Tuple

import numpy as np
import pandas as pd
from fastapi.responses import JSONResponse

from uiwiz.element import Element

LIB_PATH = Path(__file__).parent / "aggrid-community.min.js"
CSS_PATH = Path(__file__).parent / "aggridtheme.css"
JS_PATH = Path(__file__).parent / "aggrid.js"


class Aggrid(Element, extensions=[CSS_PATH, LIB_PATH, JS_PATH]):
    class OPTIONS(str, Enum):
        autoSizeColumn = "autoSizeAll"
        fitColumnContent = "sizeToFit"

    _classes: str = "ag-theme-quartz ag-theme-uiwiz w-full"

    def __init__(self, df: pd.DataFrame, column_option: OPTIONS = OPTIONS.autoSizeColumn) -> None:
        super().__init__("div")
        self.classes(Aggrid._classes)

        cols, rows = Aggrid.create_cols_and_rows(df)

        self.attributes["hx-ext"] = "hx-aggrid"
        self.attributes.__setitem__("hx-aggrid-cols", cols, False)
        self.attributes.__setitem__("hx-aggrid-rows", rows, False)
        self.attributes["hx-aggrid"] = "/data"

    @staticmethod
    def create_cols_and_rows(df: pd.DataFrame, escape: bool = True) -> Tuple[list[Any], list[Any]]:
        cols = []
        rows = []

        if df is not None:
            df = df.replace({np.nan: None})
            cols = [{"field": item} for item in df.columns.to_list()]
            rows = df.to_dict("records")
            if escape:
                cols = html.escape(json.dumps(cols))
                rows = html.escape(json.dumps(rows, default=str))

        return cols, rows

    @staticmethod
    def response(df: pd.DataFrame, headers: dict[str, str] = {}) -> JSONResponse:
        _headers = {"HX-Trigger": "aggridUpdate"} | headers
        cols, rows = Aggrid.create_cols_and_rows(df, False)
        return JSONResponse({"cols": cols, "rows": rows}, headers=_headers)



from typing import Optional

from uiwiz.elements.extensions.on_event import OnEvent


class TextArea(OnEvent):
    root_class: str = "textarea "
    root_size: str = "textarea-{size}"
    _classes: str = "textarea-bordered w-full"

    def __init__(
        self,
        name: Optional[str] = None,
        value: Optional[str] = None,
        placeholder: Optional[str] = None,
    ) -> None:
        """TextArea
        
        This element is used for text area inputs

        .. code-block:: python
            from uiwiz import ui

            ui.textarea("textarea", "Hello World", "Type something here"):

        :param name: name of the text area input
        :param value: default value of the text area input
        :param placeholder: placeholder of the text area input
        """
        super().__init__("textarea")
        self.attributes["name"] = name
        if placeholder:
            self.placeholder = placeholder
        if value:
            self.content = value
        self.classes(TextArea._classes)
        self.size(self._size)

    @property
    def placeholder(self) -> Optional[str]:
        return self.attributes["placeholder"]

    @placeholder.setter
    def placeholder(self, value: str):
        self.attributes["placeholder"] = value
        return self



from typing import Callable, Union

from uiwiz.element import Element


class Link(Element):
    root_class: str = "link "
    _classes: str = "link-hover"

    def __init__(self, text: str, link: Union[Callable[[], str], str]) -> None:
        super().__init__("a")
        self.content = text
        self.link = link
        if isinstance(link, str):
            self.attributes["href"] = link
        self.classes(Link._classes)

    def before_render(self):
        if callable(self.link):
            self.attributes["href"] = self.link()
            return
        self.attributes["href"] = self.link



from __future__ import annotations

import html
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

from typing_extensions import Self

from uiwiz.element_types import ELEMENT_SIZE, ELEMENT_TYPES, VOID_ELEMENTS
from uiwiz.event import FUNC_TYPE, TARGET_TYPE, Event
from uiwiz.frame import Frame
from uiwiz.shared import fetch_route, register_resource, route_exists


class _Attributes(dict):
    def __setitem__(self, key: Any, value: Any, escape: bool = True) -> None:
        if escape:
            if isinstance(value, str):
                value = html.escape(value)
            elif isinstance(value, bool):
                value = str(value).lower()
        super().__setitem__(key, value)


class Element:
    def __init__(
        self,
        tag: ELEMENT_TYPES = "div",
        content: str = "",
        render_html: bool = True,
        oob: bool = False,
    ) -> None:
        """Element

        Represents an HTML element. This class is used to create HTML elements.

        It is possible to create a custom element by subclassing this class.

        Content of any element will be escaped by default. If you want to render
        the content as HTML, you can use the ui.html(content="&lt;div&gt;html content&lt;/div&gt;") element.

        The rendering of the element is done by calling the render method.
        If some work needs to be done before or after rendering, the before_render
        and after_render methods can be overridden.

        The element attributes can be accessed and modified through the attributes property.
        The content of an attribute can either be a string or a callable that returns a string.

        Example:
        .. code-block:: python
            from uiwiz import ui, UiwizApp

            app = UiwizApp()

            @app.ui("/")
            async def home():
                ui.element().classes("flex")
                with ui.element().classes("relative") as container:
                    ui.element("h1", "Hello World")

                ui.element().attributes["id"] = "my-id"
                ui.element().attributes["id"] = lambda: "my-id" # <- This is also possible

        
        :param tag: The tag of the element type
        :param content: The content of the element
        :param render_html: If the element should be rendered
        :param oob: If the element should be out of band swap
        """
        self.stack = Frame.get_stack()
        if hasattr(self.__class__, "extensions") and self.__class__.extensions:
            for extension in self.__class__.extensions:
                self.stack.add_extension(self.__class__, extension)
        self.attributes: dict[str, Union[str, Callable[[], str]]] = _Attributes()
        self.attributes["id"] = self.stack.get_id()
        self.stack.id_count += 1
        self.tag: str = tag
        self._size: str = "md"

        self.event: Event = {}
        self.parent_element: Element = None
        self.children: list[Element] = []
        self.script: Optional[str] = None
        self.render_html: bool = render_html
        self.target: Optional[str] = None

        self.__content__: str = ""
        self.content: str = content
        self.oob: bool = oob

        self.classes()

        if self.oob:
            self.attributes["hx-swap-oob"] = "true"
            self.stack.oob_elements.append(self)

        if self.stack.root_element is None:
            self.stack.root_element = self
            self.stack.current_element = self
        else:
            self.stack.current_element.children.append(self)
            self.parent_element = self.stack.current_element

    def __enter__(self):
        self.stack.current_element = self
        return self

    def __exit__(self, *_):
        self.stack.current_element = self.parent_element

    def __init_subclass__(cls, extensions: List[Path] = None, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls.extensions = extensions
        if extensions:
            for extension in extensions:
                register_resource(f"{cls.__name__}/{extension.name}", extension)

    @property
    def id(self):
        return self.attributes.get("id")

    @property
    def name(self):
        return self.attributes.get("name")

    @property
    def value(self):
        return self.attributes.get("value")

    @value.setter
    def value(self, value):
        self.attributes["value"] = value

    @property
    def content(self) -> str:
        return self.__content__

    @content.setter
    def content(self, content):
        self.__content__ = html.escape(str(content))

    @property
    def is_void_element(self) -> bool:
        if self.tag in VOID_ELEMENTS:
            return True
        return False

    def get_classes(self) -> str:
        return self.attributes["class"]

    def classes(self, input: str = "") -> Self:
        """
        Set tailwind classes for the element.

        :param input: The tailwind classes to apply to the element.
        :return: The current instance of the element.
        """

        clazz = (
            getattr(self, "__root_class__", "")
            if hasattr(self, "__root_class__")
            else getattr(self.__class__, "root_class", "")
        )
        if clazz == "":
            clazz = input
        elif input:
            clazz += f" {input}"
        if clazz:
            self.attributes["class"] = clazz
            self.size(self._size)
        return self

    def size(self, size: ELEMENT_SIZE) -> Self:
        """
        Set the size of the element.

        :param size: The size of the element.
        :return: The current instance of the element.
        """
        format = getattr(self.__class__, "root_size", "")
        if format:
            old_size = format.format(size=self._size)
            if old_size in self.attributes["class"]:
                self.attributes["class"] = self.attributes["class"].replace(
                    f"{old_size}", f"{format.format(size=size)}"
                )
            else:
                clazz = self.attributes["class"]
                if clazz == "":
                    self.attributes["class"] = format.format(size=size)
                else:
                    self.attributes["class"] = (
                        f"{self.attributes['class']} {format.format(size=size)}"
                    )
            self._size = size
        return self

    def render(self, render_script: bool = True) -> str:
        output = self.render_top_level(render_script)
        output += self.render_oob()
        return output

    def render_top_level(
        self, render_script: bool = True, render_oob: bool = False
    ) -> str:
        lst = []
        lst.append(self.render_self(render_oob=render_oob))
        if render_script:
            for script in self.stack.scripts:
                lst.append(
                    """
                    <script>
                    (function() {
                    %s
                    }());
                    </script>
                    """
                    % script
                )
        return "".join(lst)

    def render_self(self, render_oob: bool = False) -> str:
        if self.oob and render_oob is False:
            return ""

        self.before_render()
        if self.script:
            self.stack.scripts.append(self.script)
        lst = []
        if self.render_html:
            self.add_event_to_attributes()

            lst.append("<%s %s>" % (self.tag, self.__dict_to_attrs__()))
            lst.append(self.content)
            lst.extend(
                [
                    child.render_self() if child.oob is False else ""
                    for child in self.children
                ]
            )

            if not self.is_void_element:
                lst.append("</%s>" % self.tag)

        html = "".join(lst)
        return self.after_render(html)

    def render_oob(self) -> str:
        lst = []
        el: Element
        for el in self.stack.oob_elements:
            lst.append(el.render_top_level(render_oob=True))
        return "".join(lst)

    def before_render(self):
        pass

    def after_render(self, html: str) -> str:
        return html

    def add_event_to_attributes(self) -> None:
        if self.event == {}:
            return None

        self.attributes["hx-target"] = self.get_target(self.event.get("target"))
        self.attributes["hx-swap"] = (
            self.event.get("swap")
            if self.event.get("swap") is not None
            else "outerHTML"
        )

        self.attributes["hx-post"] = self.__get_endpoint__(self.event["func"])
        self.attributes["hx-trigger"] = self.event.get("trigger")

        if vals := self.event.get("vals"):
            self.attributes["hx-vals"] = vals
        if include := self.event.get("include"):
            self.attributes["hx-include"] = include
        if hx_encoding := self.event.get("hx-encoding"):
            self.attributes["hx-encoding"] = hx_encoding
        else:
            self.attributes["hx-ext"] = "json-enc"

    def __get_endpoint__(self, func: FUNC_TYPE) -> str:
        if isinstance(func, str):
            return func

        endpoint: Optional[str] = fetch_route(func)
        if endpoint:
            if params := self.event.get("params"):
                return endpoint.format(**params)
            return endpoint

        endpoint = f"/_uiwiz/hash/{func.__hash__()}"
        if not route_exists(endpoint):
            self.stack.app.ui(endpoint)(func)
        return endpoint

    def get_target(self, target: TARGET_TYPE) -> str:
        _target = "this"
        if target is None:
            return _target

        if isinstance(target, Callable):
            return "#%s" % str(target())

        if isinstance(target, Element):
            return "#%s" % target.id

        if target != "this" and "#" not in target:
            return "#" + target

        return target

    def render_ext(self, lst_ext: list[str]) -> Tuple[str, str]:
        lst_js = []
        lst_css = []
        for lib in lst_ext:
            if lib.endswith("css"):
                css = '<link href="%s" rel="stylesheet" type="text/css" />' % lib
                if css not in lst_css:
                    lst_css.insert(0, css)
            elif lib.endswith("js"):
                js = '<script src="%s"></script>' % lib
                if js not in lst_js:
                    lst_js.append(js)
            else:
                raise Exception("lib type not supported, supported types css, js")
        return "".join(lst_js), "".join(lst_css)

    def __str__(self) -> str:
        return self.render()

    def set_frame(self, frame: Frame):
        self.stack = frame
        for child in self.children:
            child.set_frame(frame)

    def set_frame_and_root(self) -> None:
        self.set_frame(Frame.get_stack())
        self.stack.root_element = self

    def __dict_to_attrs__(self):
        ATTR_NO_VALUE = object()
        return " ".join(
            (
                key
                if value is ATTR_NO_VALUE
                else (
                    '%s="%s"' % (key, value())
                    if callable(value)
                    else '%s="%s"' % (key, value)
                )
            )
            for key, value in self.attributes.items()
        )



from typing import Any, Optional

from uiwiz.element import Element


class HiddenInput(Element):
    _classes: str = ""

    def __init__(
        self,
        name: Optional[str] = None,
        value: Any = None,
    ) -> None:
        """HiddenInput

        This element is used for hidden input data

        :param name: Used in the attributes name sent back in json
        :param value: The value to send back
        """
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["value"] = value
        self.attributes["type"] = "hidden"



from typing import TYPE_CHECKING, Callable, Literal, Optional, TypedDict, Union

if TYPE_CHECKING:
    from uiwiz.element import Element

TRIGGER_COMBINATIONS = Literal[
    "keydown[ctrlKey&&key=='Enter'], keydown[metaKey&&key=='Enter']",
    "keydown[key=='Enter']",
    "keydown[ctrlKey&&key=='Enter']",
    "keydown[metaKey&&key=='Enter']",
    "keydown[shiftKey&&key=='Enter']",
    "keydown[altKey&&key=='Enter']",
]

ON_EVENTS = Union[
    Literal[
        "input",
        "change",
        "click",
        "focus",
        "blur",
        "submit",
        "reset",
        "keydown",
        "keyup",
        "keypress",
        "scroll",
        "copy",
        "cut",
        "paste",
    ],
    TRIGGER_COMBINATIONS,
]

# HTMX Swap Events
# https://htmx.org/attributes/hx-swap/
SWAP_EVENTS = Optional[
    Literal[
        "innerHTML",
        "outerHTML",
        "textContent",
        "beforebegin",
        "afterbegin",
        "beforeend",
        "afterend",
        "delete",
        "none",
    ]
]

TARGET_TYPE = Optional[Union[Callable[[], str], str, "Element", None]]
FUNC_TYPE = Union[Callable, str]


class Event(TypedDict):
    func: FUNC_TYPE
    trigger: ON_EVENTS
    target: TARGET_TYPE
    swap: SWAP_EVENTS
    include: Optional[str]
    vals: Optional[str]
    params: Optional[dict]



from uiwiz.element import Element


class Col(Element):
    root_class: str = "flex flex-col {item_position} {gap} {padding} "

    def __init__(
        self,
        item_position: str = "items-start",
        gap: str = "gap-4",
        padding: str = "p-4",
    ) -> None:
        """Col

        Align children elements horizontally.

        :param item_position: The position of the items in the column. Items-start by default and will align the items to the left. items-end will align the items to the right.
        :param gap: The gap between the items in the column.
        :param padding: The padding of the column.
        """
        super().__init__()
        self.__root_class__ = Col.root_class.format(
            item_position=item_position, gap=gap, padding=padding
        )



from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.checkbox import Checkbox
from uiwiz.elements.html import Html
from uiwiz.elements.label import Label
from uiwiz.svg.svg_handler import get_svg


class DrawerSetup(Element):
    _classes: str = "menu p-4 w-80 min-h-full bg-base-200 text-base-content"

    def __init__(self) -> None:
        super().__init__("ul")
        self.classes(DrawerSetup._classes)


class DrawerSide(Element):
    _classes: str = "drawer-side z-50"

    def __init__(self, drawer_toggle: Checkbox) -> None:
        super().__init__()
        self.classes(DrawerSide._classes)
        self.drawer_toggle = drawer_toggle

    def __enter__(self):
        super().__enter__()
        close_label = Label("", self.drawer_toggle)
        close_label.attributes["class"] = "drawer-overlay"
        close_label.attributes["aria-label"] = "close sidebar"
        self.setup = DrawerSetup()
        self.setup.__enter__()

    def __exit__(self, *_):
        return super().__exit__(*_)


class DrawerContent(Element):
    _classes: str = "drawer-content flex flex-col"

    def __init__(self) -> None:
        super().__init__()
        self.classes(DrawerContent._classes)


class Drawer(Element):
    _classes: str = "drawer bg-base-100 min-h-screen h-full"

    def __init__(self, always_open: bool = False, right: bool = False) -> None:
        """Drawer

        A drawer is a panel that slides in from the side of the screen. It can be used to display additional content or controls.

        .. code-block:: python
            from uiwiz import ui

            with ui.drawer() as drawer:
                with drawer.drawer_content():
                    with ui.nav():
                        drawer.drawer_button()

                    with ui.col():
                        ui.label("test1")
                        ui.button("Click me")
                        ui.label("test1")

                with drawer.drawer_side():
                    with ui.element("li"):
                        ui.link("Drawer-1", "/")

        :param always_open: Always open the drawer until the screen size is too small.
        :param right: Open the drawer from the right side of the screen.
        """
        super().__init__()
        self.classes(Drawer._classes)
        self.__drawer_button_menu__: Optional[Element] = None

        if always_open:
            self.classes(self.attributes["class"] + " lg:drawer-open")
        if right:
            self.classes(self.attributes["class"] + " drawer-end")

        with self:
            self.drawer_toggle = Checkbox("")
            self.drawer_toggle.attributes["class"] = "drawer-toggle"

    def drawer_content(self) -> Element:
        self.drawer_content = DrawerContent()
        return self.drawer_content

    def drawer_side(self) -> Element:
        self.side = DrawerSide(self.drawer_toggle)
        return self.side

    def drawer_button(self) -> Element:
        self.__drawer_button_menu__ = Element("label").classes("btn drawer-button")
        self.__drawer_button_menu__.attributes["for"] = self.drawer_toggle.id
        with self.__drawer_button_menu__:
            Html(get_svg("menu"))



from fastapi import Request

from uiwiz.element import Element
from uiwiz.event import ON_EVENTS, SWAP_EVENTS


class Bindable(Element):
    def bind_text_from(
        self,
        element: Element,
        trigger: ON_EVENTS = "input delay:20ms",
        swap: SWAP_EVENTS = "outerHTML",
    ):
        """
        Bind the text of this element to data of another element.
        Requires the other element to have a name attribute.
        :param element: Element to bind to
        :param trigger: Event trigger to bind to
        """
        assert element.attributes.get("name") is not None

        async def bind_value(request: Request):
            data = await request.json()
            self.set_frame_and_root()

            input_data = data[element.name]

            self.content = str(input_data)

        bind_value.__name__ = f"{bind_value.__name__}_{element.id}"

        element.event = {
            "target": f"#{self.id}",
            "func": bind_value,
            "trigger": trigger,
            "swap": swap,
        }

        return self



from typing import Any, List, Literal, Optional, TypedDict, Union


class SqlOptions(TypedDict):
    tables: list[str]
    columns: list[str]


class AceOptions(TypedDict):
    animated_scroll: Optional[bool]
    auto_scroll_editor_into_view: Optional[bool]
    behaviours_enabled: Optional[bool]
    copy_with_empty_selection: Optional[bool]
    cursor_style: Optional[Literal["ace", "slim", "smooth", "wide"]]
    custom_scrollbar: Optional[bool]
    display_indent_guides: Optional[bool]
    drag_delay: Optional[int]
    drag_enabled: Optional[bool]
    enable_auto_indent: Optional[bool]
    enable_basic_autocompletion: Optional[Union[bool, List[Any]]]
    enable_codeLens: Optional[bool]
    enable_keyboard_accessibility: Optional[bool]
    enable_live_autocompletion: Optional[Union[bool, List[Any]]]
    enable_mobile_menu: Optional[bool]
    enable_multiselect: Optional[bool]
    enable_snippets: Optional[bool]
    fade_fold_widgets: Optional[bool]
    first_line_number: Optional[int]
    fixed_width_gutter: Optional[bool]
    focus_timeout: Optional[int]
    fold_style: Optional[Literal["markbegin", "markbeginend", "manual"]]
    font_family: Optional[str]
    font_size: Optional[Union[str, int]]
    h_scroll_bar_always_visible: Optional[bool]
    has_css_transforms: Optional[bool]
    highlight_active_line: Optional[bool]
    highlight_gutter_line: Optional[bool]
    highlight_indent_guides: Optional[bool]
    highlight_selected_word: Optional[bool]
    indented_soft_wrap: Optional[bool]
    keyboard_handler: Optional[str]
    live_autocompletion_delay: Optional[int]
    live_autocompletion_threshold: Optional[int]
    max_lines: Optional[int]
    max_pixel_height: Optional[int]
    merge_undo_deltas: Optional[Union[bool, Literal["always"]]]
    min_lines: Optional[int]
    mode: Optional[str]
    navigate_within_soft_tabs: Optional[bool]
    newLine_mode: Optional[Literal["auto", "windows", "unix"]]
    overwrite: Optional[bool]
    placeholder: Optional[str]
    print_margin: Optional[Union[int, bool]]
    print_margin_column: Optional[int]
    read_only: Optional[bool]
    relative_line_numbers: Optional[bool]
    scroll_past_end: Optional[int]
    scroll_speed: Optional[int]
    selection_style: Optional[Literal["line", "text", "fullLine", "screenLine"]]
    show_fold_widgets: Optional[bool]
    show_folded_annotations: Optional[bool]
    show_gutter: Optional[bool]
    show_invisibles: Optional[bool]
    show_line_numbers: Optional[bool]
    show_print_margin: Optional[bool]
    tab_size: Optional[int]
    text_input_aria_label: Optional[str]
    theme: Optional[str]
    tooltip_follows_mouse: Optional[bool]
    use_resize_observer: Optional[bool]
    use_soft_tabs: Optional[bool]
    use_svg_gutter_icons: Optional[bool]
    use_worker: Optional[bool]
    v_scroll_bar_always_visible: Optional[bool]
    value: Optional[str]
    wrap: Optional[Union[int, float, bool, Literal["off", "free", "printmargin"]]]
    wrap_behaviours_enabled: Optional[bool]
    wrap_method: Optional[Literal["code", "text", "auto"]]



from typing import Literal

from anyio import Path

_type = Literal["info", "error", "success", "warning", "menu", "copy"]


def get_svg(svg: _type) -> str:
    if not svg:
        raise ValueError("Value cannot be None or an empty string")
    path = Path(__file__).parent / (svg + ".svg")
    with open(path, "r") as f:
        return f.read()



import json
from pathlib import Path

from fastapi.responses import JSONResponse

from uiwiz.element import Element

LIB_PATH = Path(__file__).parent / "echart.min.js"
JS_PATH = Path(__file__).parent / "echart.js"


class EChart(Element, extensions=[LIB_PATH, JS_PATH]):
    name: str = "data-wz-echart"

    def __init__(self, options: dict, height: str = "h-80") -> None:
        """EChart element
        
        See https://echarts.apache.org/examples/en/index.html for examples
        on how to use ECharts

        Example usage:
        .. code-block:: python
                
                from uiwiz import ui
    
                options = {
                    "title": {"text": "ECharts entry example"},
                    "tooltip": {},
                    "legend": {"data": ["Sales"]},
                    "xAxis": {"data": ["shirt", "cardign", "chiffon shirt", "pants", "heels", "socks"]},
                    "yAxis": {},
                    "series": [{"name": "Sales", "type": "bar", "data": [5, 20, 36, 10, 10, 20]}],
                }
    
                ui.echart(options)

        :param options: EChart options
        :param height: Height of the chart container
        """
        with Element() as container:
            container.classes(f"flex justify-center relative overflow-hidden items-center {height}")
            super().__init__()
        
        self.attributes[EChart.name] = EChart.name
        self.attributes[f"{EChart.name}-options"] = json.dumps(options)
        self.attributes["hx-ext"] = EChart.name
        self.classes("w-full h-full")

    def container_classes(self, input: str) -> "EChart":
        self.parent_element.classes(input)
        return self

    @staticmethod
    def response(data: dict, headers: dict[str, str] = {}) -> JSONResponse:
        _headers = {"HX-Trigger": json.dumps({"uiwizUpdateEChart": data})} | headers
        return JSONResponse(data, headers=_headers)



from uiwiz.elements.extensions.on_event import OnEvent
from uiwiz.event import FUNC_TYPE, ON_EVENTS, SWAP_EVENTS, TARGET_TYPE


class Upload(OnEvent):
    root_class: str = "file-input"
    root_size: str = "file-input-{size}"
    _classes: str = "file-input-bordered"

    def __init__(
        self,
        name: str,
    ) -> "Upload":
        """Upload

        This element is used for file uploads

        .. code-block:: python
            from uiwiz import ui
            from fastapi import UploadFile

            @app.ui("/upload/endpoint")
            async def handle_upload(file: UploadFile):
                file_output = await file.read()

            ui.upload("file").on_upload(on_upload=handle_upload, swap="none")
        
        """
        super().__init__("input")
        self.attributes["type"] = "file"
        self.attributes["name"] = name
        self._size = "sm"
        self.classes(Upload._classes)

    def on_upload(
        self,
        on_upload: FUNC_TYPE,
        target: TARGET_TYPE = None,
        trigger: ON_EVENTS = "change",
        swap: SWAP_EVENTS = None,
    ) -> "Upload":
        """
        :param on_upload: The function to call when the upload event is triggered or the endpoint to call
        :param target: The target to swap the response to
        :param trigger: The event to trigger the function
        :param swap: The swap event to use
        """
        self.event = {
            "func": on_upload,
            "trigger": trigger,
            "target": target,
            "swap": swap,
            "hx-encoding": "multipart/form-data",
        }

        return self



import json
from pathlib import Path
from typing import Literal, Optional

import humps

from uiwiz.element import Element
from uiwiz.elements.ace import AceOptions, SqlOptions
from uiwiz.elements.form import Form

LIB_PATH = Path(__file__).parent / "ace.min.js"
MODE_PYTHON = Path(__file__).parent / "mode-python.js"
SNIPPETS_PYTHON = Path(__file__).parent / "snippets/python.js"
SNIPPETS_SQL = Path(__file__).parent / "snippets/sql.js"
MODE_SQL = Path(__file__).parent / "mode-sql.js"
LIB_LANG_TOOL_PATH = Path(__file__).parent / "ace-lang-tool.min.js"
CSS_PATH = Path(__file__).parent / "acetheme.css"
JS_PATH = Path(__file__).parent / "ace.js"


class Ace(
    Element,
    extensions=[
        CSS_PATH,
        LIB_PATH,
        LIB_LANG_TOOL_PATH,
        MODE_PYTHON,
        SNIPPETS_PYTHON,
        SNIPPETS_SQL,
        MODE_SQL,
        JS_PATH,
    ],
):
    default_options = AceOptions(
        enable_basic_autocompletion=True,
        enable_live_autocompletion=True,
        enable_snippets=True,
        selection_style="text",
        highlight_active_line=False,
        highlight_gutter_line=False,
        print_margin_column=False,
        show_print_margin=False,
    )

    root_class: str = "ace-editor "
    _classes: str = "rounded-md w-full h-56 resize-y"

    def __init__(
        self,
        name: Optional[str] = None,
        content: str = None,
        lang: Literal["sql", "python"] = "python",
        sql_options: Optional[SqlOptions] = None,
        ace_options: Optional[AceOptions] = None,
    ) -> None:
        """Ace Editor

        Use the Ace Editor to edit code in a textarea element.

        Example:
        .. code-block:: python
            from uiwiz import ui

            ui.ace(name="editor")

        :param name: The name of the textarea element to be submitted with a form
        :param content: The initial content of the editor
        :param lang: The language mode to use. One of "sql" or "python"
        :param sql_options: Options for the SQL language mode. Tables and columns to be used for autocompletion
        :param ace_options: Options for the Ace Editor
        """
        hidden_text = Element("textarea")
        hidden_text.attributes["hidden"] = "true"
        hidden_text.attributes["name"] = name
        hidden_text.content = content or ""

        super().__init__()
        hidden_text.attributes["hx-ace-editor-id"] = self.id

        self.classes(Ace._classes)
        self.attributes["hx-ace-editor-lang"] = lang
        self.attributes["hx-ace-editor-hidden-input"] = hidden_text.id
        self.attributes["hx-ace-editor-form"] = self.find_parent_form()
        self.attributes["hx-ace-editor-content"] = content

        self.attributes["hx-ace-editor-options"] = (
            json.dumps(humps.camelize(ace_options)) if ace_options else json.dumps(humps.camelize(Ace.default_options))
        )
        self.attributes["hx-ace-editor-sql-options"] = json.dumps(sql_options) if sql_options else "{}"

    def find_parent_form(self) -> Optional[str]:
        parent = self.parent_element
        while parent:
            if isinstance(parent, Form):
                return parent.id
            parent = parent.parent_element
        return None



from pathlib import Path

import markdown2

from uiwiz.element import Element
from uiwiz.elements.html import Html

MARKDOWN = Path(__file__).parent / "markdown.css"
CODE_HIGHLIGHT = Path(__file__).parent / "codehighlight.css"


class Markdown(Element, extensions=[MARKDOWN, CODE_HIGHLIGHT]):
    def __init__(self, content="", extras=["fenced-code-blocks", "markdown-in-html"]) -> None:
        super().__init__()
        self.classes("markdown-body")
        self.extras = extras
        with self:
            _content = (
                markdown2.markdown(content, extras=self.extras)
                .replace("<ul>", '<ul class="list-disc">')
                .replace("<ol>", '<ol class="list-decimal">')
            )

            self.markdown = Html(content=_content)



import typing
from typing import Literal, Optional

from uiwiz import ui
from uiwiz.asgi_request_middleware import get_request
from uiwiz.element import Element

THEMES = Literal[
    "light",
    "dark",
    "nord",
    "cupcake",
    "pastel",
    "bumblebee",
    "emerald",
    "corporate",
    "synthwave",
    "retro",
    "cyberpunk",
    "valentine",
    "halloween",
    "garden",
    "forest",
    "aqua",
    "lofi",
    "fantasy",
    "wireframe",
    "black",
    "luxury",
    "dracula",
    "cmyk",
    "autumn",
    "business",
    "acid",
    "lemonade",
    "night",
    "coffee",
    "winter",
    "dim",
    "sunset",
]


class ThemeSelector(Element):
    def __init__(self, themes: Optional[THEMES] = None) -> None:
        super().__init__()
        self.render_html = False
        self.themes = list(typing.get_args(THEMES))
        if themes:
            self.themes = themes

        placeholder = "Theme"
        if theme := get_request().cookies.get("data-theme"):
            placeholder = theme

        self.theme_selector = ui.dropdown("theme-selector", self.themes, placeholder)
        self.theme_selector.classes("min-w-32")
        self.setup_listener()

    def setup_listener(self):
        self.script = f"""
function selectTheme(value) {{
    console.log(value);
    element = document.getElementById("html");
    element.setAttribute("data-theme", value);
    document.cookie = `data-theme=${{value}}; Path=/`; 
}}

document.getElementById("{self.theme_selector.id}").addEventListener('change', function() {{
    selectTheme(this.value);
}});
"""



from uiwiz.element import Element


class Row(Element):
    root_class: str = "flex flex-row {wrap} {item_position} {gap} {padding} "

    def __init__(
        self,
        wrap: str = "flex-wrap",
        item_position: str = "items-start",
        gap: str = "gap-4",
        padding: str = "",
    ) -> None:
        """Row
        
        Align children elements vertically.

        .. code-block:: python
            from uiwiz import ui

            with ui.row():
                ui.label("Hello")
                ui.label("World")

        :param wrap: The wrap of the row. Flex-wrap by default and will wrap the items to the next row if they don't fit.
        :param item_position: The position of the items in the row. Items-start by default and will align the items to the left. items-end will align the items to the right.
        :param gap: The gap between the items in the row.
        :param padding: The padding of the row.
        """
        super().__init__()
        self.__root_class__ = Row.root_class.format(
            wrap=wrap, item_position=item_position, gap=gap, padding=padding
        )



from uiwiz.element import Element


class Divider(Element):
    root_class = "divider "
    _classes_hor = "divider-horizontal"

    def __init__(self, text: str = "") -> None:
        """Divider

        Display a divider line between elements.

        :param text: The text to display in the divider.
        """
        super().__init__("div")
        self.content = text

    def horizontal(self) -> "Divider":
        self.classes(Divider._classes_hor)
        return self



import json
from typing import Optional

from uiwiz import ui
from uiwiz.element import Element
from uiwiz.svg.svg_handler import _type, get_svg


class Toast(Element):
    root_class: str = "alert w-full z-50 "

    def __init__(self, message: str = "", svg: _type = None) -> None:
        """Toast

        Display a toast message on the client side.
        Can be used anywhere in a @page_router.ui("<path>") or @app.ui("<path>")

        .. code-block:: python
            @app.ui("/some/path")
            async def some_path():
                ui.toast("This is a toast message").info()

        To persist the toast message, set auto_close to False

        .. code-block:: python
            @app.ui("/some/path")
            async def some_path():
                ui.toast("This is a toast message").info().set_auto_close(False)

        :param message: The message to display
        :param svg: The svg to display. One of "info", "warning", "success", "error", "menu"
        
        """
        self.inner_class = Toast.root_class
        super().__init__(tag="div", oob=True)
        self.attributes["id"] = "toast"
        self.attributes["hx-swap-oob"] = "afterbegin"
        self._svg: _type = svg
        self.message: str = message
        self.context_manager_used: bool = False
        self.inner_element: Optional[Element] = None
        self._auto_close: bool = True

    @property
    def auto_close(self) -> bool:
        return self._auto_close
    
    @auto_close.setter
    def auto_close(self, auto_close: bool) -> None:
        self._auto_close = auto_close
    
    def set_auto_close(self, auto_close: bool) -> "Toast":
        """Set auto close
        
        :param auto_close: True or False. Auto close False will keep the toast open until the user closes it
        """
        self.auto_close = auto_close
        return self

    def before_render(self):
        with self:
            with self.inner_element.classes(self.inner_class + " relative pr-16"):
                if self._svg:
                    svg = ui.html(get_svg(self._svg))
                    self.inner_element.children.remove(svg)
                    self.inner_element.children.insert(0, svg)
                Element("span", content=self.message)
                self.inner_element.attributes["hx-toast-data"] = json.dumps({"autoClose": self._auto_close})
                btn = None
                self.inner_element.attributes["hx-toast-delete-button"] = lambda: btn.id if btn else ""
                if not self._auto_close:
                    btn = ui.button("✕").classes("btn btn-sm btn-circle btn-ghost absolute right-2 top")

    def __enter__(self):
        super().__enter__()
        if self.inner_element is None:
            self.inner_element = Element().classes(self.inner_class)
        self.inner_element.__enter__()
        return self

    def svg(self, svg: _type) -> "Toast":
        self._svg = svg
        return self

    def classes(self, input: str = "") -> "Toast":
        self.inner_class = self.inner_class + input
        return self

    def info(self) -> "Toast":
        self.classes("alert")
        self.svg("info")
        return self

    def warning(self) -> "Toast":
        self.classes("alert-warning")
        self.svg("warning")
        return self

    def success(self) -> "Toast":
        self.classes("alert-success")
        self.svg("success")
        return self

    def error(self) -> "Toast":
        self.classes("alert-error")
        self.svg("error")
        self.auto_close = False
        return self


from typing import Callable, List, Optional, get_type_hints

import numpy as np
import pandas as pd
from pydantic import BaseModel

from uiwiz.element import Element
from uiwiz.element_types import ELEMENT_SIZE
from uiwiz.elements.button import Button
from uiwiz.elements.form import Form
from uiwiz.event import FUNC_TYPE
from uiwiz.models.model_handler import ModelForm


class ModelFormRender(ModelForm):
    def render_model(self, *args, **kwargs) -> Form:
        self.button = Button("Save")
        self.button.render_html = False


class Table(Element):
    root_size: str = "table-{size}"
    _classes_container: str = "w-full overflow-x-auto rounded-box "
    _classes_table: str = (
        "table table-zebra table-auto bg-base-300 overflow-scroll w-full whitespace-nowrap pr-4 pt-2 pb-2"
    )

    def __init__(self, data: List[BaseModel], id_column_name: Optional[str] = None) -> None:
        """
        Creates a table from a list of pydantic models

        :param data: A list of pydantic models
        :param id_column_name: The name of the Pydantic attribute to be used with the path param endpoint. An endpoint like /path/{id} should have a attribute "id" on the class
        :return: The current instance of the element.
        """
        super().__init__()
        self.classes(Table._classes_container)

        self.schema = []
        if data:
            self.schema = list(data[0].model_fields.keys())

        self.data = data
        self.did_render: bool = False
        self.edit: Optional[FUNC_TYPE] = None
        self.delete: Optional[FUNC_TYPE] = None
        self.create: Optional[FUNC_TYPE] = None
        self.id_column_name: Optional[str] = id_column_name

    def set_border(self, border_classes: str = "border border-base-content") -> "Table":
        """
        Set the border classes for the table

        :param border_classes: The border classes to set
        :return: The current instance of the element.
        """
        self.classes(Table._classes_container + border_classes)
        return self

    def edit_row(self, edit: FUNC_TYPE) -> "Table":
        """
        Enable editing functionality for a table row.

        This method assigns the endpoint to handle the "Edit" action for rows
        in the table. It requires that `id_column_name` be set to identify rows uniquely.

        :param edit: The endpoint that is triggered when the "Edit" action is performed.
            The callback should accept parameters required to handle the editing logic.
        :type edit: FUNC_TYPE
        :return: Returns the `Table` instance with the "Edit" functionality enabled.
        :rtype: Table
        :raises ValueError: If `id_column_name` is not set, as it is required to identify rows
            for the "Edit" operation.

        Example:
            >>> table = Table(data, id_column_name="id").edit_row(edit=edit_endpoint)
        """
        if self.id_column_name is None:
            raise ValueError("When using edit id_column_name is required")
        self.edit = edit
        return self

    def delete_row(self, delete: FUNC_TYPE) -> "Table":
        """
        Enable deletion functionality for a table row.

        This method assigns the endpoint to handle the "Delete" action for rows
        in the table. It requires that `id_column_name` be set to identify rows uniquely.

        :param delete: The endpoint that is triggered when the "Delete" action is performed.
            The callback should accept parameters required to handle the deletion logic.
        :type delete: FUNC_TYPE
        :return: Returns the `Table` instance with the "Delete" functionality enabled.
        :rtype: Table
        :raises ValueError: If `id_column_name` is not set, as it is required to identify rows
            for the "Delete" operation.

        Example:
            >>> table = Table(data, id_column_name="id").delete_row(delete=delete_endpoint)
        """
        if self.id_column_name is None:
            raise ValueError("When using delete id_column_name is required")
        self.delete = delete
        return self

    def create_row(self, create: FUNC_TYPE) -> "Table":
        """
        Enable creation functionality for a table row.

        This method assigns the endpoint to handle the "Create" action for a row
        in the table.

        :param create: The endpoint that is triggered when the "Delete" action is performed.
        :type create: FUNC_TYPE
        :return: Returns the `Table` instance with the "Delete" functionality enabled.
        :rtype: Table

        Example:
            >>> table = Table(data).create_row(create=create_endpoint)
        """
        self.create = create
        return self

    @classmethod
    def render_edit_row(
        cls,
        model: BaseModel,
        id_column_name: str,
        save: FUNC_TYPE,
        cancel: FUNC_TYPE,
        size: ELEMENT_SIZE = "sm",
        **kwargs,
    ) -> Element:
        """
        Render a table row with inputs and cancel/save button

        :param model: The Pydantic model to render
        :param id_column_name: The column that should be used with edit or delete as path param
        :param save: The endpoint to call when save button is clicked
        :param cancel: The endpoint to call when cancel is clicked
        :param size: The size of the inputs
        :return: The tr element container
        """
        with Element("tr") as container:
            rendere = ModelFormRender(model, size=size)
            hints = get_type_hints(model, include_extras=True)
            for key, field_type in hints.items():
                with Element("td"):
                    rendere.render_model_attributes(key, field_type, **kwargs)

            Table.__render_save_button__(container, save, cancel, id_column_name, model, size, **kwargs)
        return container

    @classmethod
    def __render_save_button__(
        cls,
        container: Element,
        save: Callable,
        cancel: Callable,
        id_column_name: str,
        model: BaseModel,
        size: ELEMENT_SIZE = "sm",
        **kwargs,
    ) -> Element:
        with Element("td").classes("flex justify-end join"):
            Button("Cancel").size(size).on(
                "click",
                cancel,
                container,
                "outerHTML",
                params={
                    id_column_name: (
                        model.__getattribute__(id_column_name)
                        if isinstance(model, BaseModel)
                        else kwargs.get(id_column_name, {}).get("value")
                    )
                },
            ).classes("btn-warning border border-base-content join-item flex-1 flex-initial").attributes[
                "hx-include"
            ] = "closest tr"
            Button("Save").size(size).on(
                "click",
                save,
                container,
                "outerHTML",
                params={
                    id_column_name: (
                        model.__getattribute__(id_column_name)
                        if isinstance(model, BaseModel)
                        else kwargs.get(id_column_name, {}).get("value")
                    )
                },
            ).classes("btn-success border border-base-content join-item flex-1 flex-initial").attributes[
                "hx-include"
            ] = "closest tr"

    def before_render(self) -> None:
        super().before_render()
        if self.did_render:
            return None

        with self:
            with Element("table").classes(Table._classes_table):
                # columns
                with Element("thead"):
                    with Element("tr"):
                        for col in self.schema:
                            Element("th", content=col)
                        if self.edit:
                            Element("th")
                # rows
                with Element("tbody") as container:
                    for row in self.data:
                        self.render_row(row, self.id_column_name, self.edit, self.delete)
        if self.create:
            Button("Add").on_click(self.create, container, swap="beforeend")
        self.did_render = True

    @classmethod
    def render_row(
        cls,
        row: BaseModel,
        id_column_name: Optional[str] = None,
        edit: Optional[FUNC_TYPE] = None,
        delete: Optional[FUNC_TYPE] = None,
        size: ELEMENT_SIZE = "sm",
    ) -> Element:
        """
        Render a table row

        :param row: The instance Pydantic model to render
        :param id_column_name: The optional column that should be used with edit or delete
        :param edit: The optional endpoint to call with the path param from id_column_name
        :param delete: The optional endpoint to call with the parh param from id_column_name
        :return: The tr element container
        """
        with Element("tr") as container:
            for item in list(row.model_fields.keys()):
                Element("td", content=row.__getattribute__(item))
            if edit and delete:
                with Element("td").classes("flex justify-end join"):
                    Button("Edit").classes("border border-base-content flex-1 join-item flex-initial").size(size).on(
                        "click",
                        edit,
                        container,
                        "outerHTML",
                        params={id_column_name: row.__getattribute__(id_column_name)},
                    )
                    Button("Delete").classes("btn-error border border-base-content join-item flex-1 flex-initial").size(
                        size
                    ).on(
                        "click",
                        delete,
                        container,
                        "outerHTML",
                        params={id_column_name: row.__getattribute__(id_column_name)},
                    ).attributes["hx-confirm"] = "Are you sure?"
            elif edit:
                with Element("td").classes("flex justify-end"):
                    Button("Edit").classes("border border-base-content flex-1 flex-initial").size(size).on(
                        "click",
                        edit,
                        container,
                        "outerHTML",
                        params={id_column_name: row.__getattribute__(id_column_name)},
                    )
            elif delete:
                with Element("td").classes("flex justify-end"):
                    Button("Delete").classes("border border-base-content flex-1 flex-initial").size(size).on(
                        "click",
                        delete,
                        container,
                        "outerHTML",
                        params={id_column_name: row.__getattribute__(id_column_name)},
                    )
        return container

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame) -> Element:
        """
        Render a pandas.DataFrame

        :param df: The DataFrame to render
        :return: The container element
        """
        df = df.replace({np.nan: "None"})

        with Element().classes(Table._classes_container) as container:
            with Element("table").classes(Table._classes_table):
                # columns
                with Element("thead"):
                    with Element("tr"):
                        for col in df.columns:
                            Element("th", content=col)
                # rows
                with Element("tbody"):
                    for _, row in df.iterrows():
                        with Element("tr"):
                            for _, val in row.items():
                                Element("td", content=val)
        return container
